# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from .exceptions import InsufficientData, MoonchartError
from .utils import (
    get_zscores,
    trim_outliers as trim_outliers_func,
    get_sharpe,
    get_rolling_sharpe,
    get_cagr,
    get_cum_returns,
    get_drawdowns,
    intraday_to_daily)

from quantrocket.moonshot import read_moonshot_csv
from quantrocket.blotter import read_pnl_csv

class DailyPerformance(object):
    """
    Class representing daily performance and derived statistics.

    Typically constructed using DailyPerformance.from_moonshot_csv.

    Parameters
    ----------
    returns : DataFrame, required
        a Dataframe of pct returns

    pnl : DataFrame, optional
        a DataFrame of pnl

    net_exposures : DataFrame, optional
        a Dataframe of net (hedged) exposure

    abs_exposures : DataFrame, optional
        a Dataframe of absolute exposure (ignoring hedging)

    total_holdings : DataFrame, optional
        a Dataframe of the number of holdings

    turnover : DataFrame, optional
        a DataFrame of turnover, that is, changes to positions

    commission_amounts : DataFrame, optional
        a DataFrame of commission amounts, in the base currency

    commissions : DataFrame, optional
        a DataFrame of commissions, as a proportion of capital

    slippages : DataFrame, optional
        a DataFrame of slippages, as a proportion of capital

    benchmark : Series, optional
        a Series of prices for a benchmark

    riskfree : float, optional
        the riskfree rate (default 0)

    compound : bool
         True for compound/geometric returns, False for arithmetic returns (default True)

    rolling_sharpe_window : int, optional
        compute rolling Sharpe over this many periods (default 200)

    trim_outliers: int or float, optional
        discard returns that are more than this many standard deviations from the mean
    """

    def __init__(
        self,
        returns,
        pnl=None,
        net_exposures=None,
        abs_exposures=None,
        total_holdings=None,
        turnover=None,
        commission_amounts=None,
        commissions=None,
        slippages=None,
        benchmark=None,
        riskfree=0,
        compound=True,
        rolling_sharpe_window=200,
        trim_outliers=None
        ):

        self.returns = returns
        if len(self.returns.index) < 2:
            raise InsufficientData(
                "Moonchart needs at least 2 dates to analyze performance, "
                "but returns DataFrame has length {0}".format(len(self.returns.index)))
        self._trim_outliers = trim_outliers
        if trim_outliers:
            self.returns = trim_outliers_func(returns, z_score=trim_outliers)
        else:
            # warn the user if there are >10-sigma returns
            z_scores = get_zscores(returns)
            max_zscore = z_scores.abs().max()
            if isinstance(returns, pd.DataFrame):
                max_zscore = max_zscore.max()
            if max_zscore > 20:
                import warnings
                warnings.warn("Found returns which are {0} standard deviations from the "
                              "mean, consider removing them with the `trim_outliers` parameter".format(
                                  round(max_zscore)))
        self.pnl = pnl
        self.net_exposures = net_exposures
        self.abs_exposures = abs_exposures
        self.total_holdings = total_holdings
        self.turnover = turnover
        self.commission_amounts = commission_amounts
        self.commissions = commissions
        self.slippages = slippages
        self.riskfree = riskfree
        self.compound = compound
        self.rolling_sharpe_window = rolling_sharpe_window
        self._benchmark_prices = benchmark
        self._benchmark_returns = None
        self._benchmark_cum_returns = None
        self._cum_returns = None
        self._cum_commissions = None
        self._cum_slippages = None
        self._cum_commission_amounts = None
        self._sharpe = None
        self._rolling_sharpe = None
        self._cagr = None
        self._drawdowns = None
        self._max_drawdown = None
        self._cum_pnl = None

    @classmethod
    def _from_moonshot_or_pnl(cls, results,
                              start_date=None,
                              end_date=None,
                              trim_outliers=None,
                              how_to_aggregate=None,
                              riskfree=0,
                              compound=True,
                              rolling_sharpe_window=200):
        """
        Creates a DailyPerformance instance from a moonshot backtest results
        DataFrame or a PNL DataFrame.
        """
        if "Time" in results.index.names:
            results = intraday_to_daily(results, how=how_to_aggregate)

        if start_date:
            results = results[results.index.get_level_values("Date") >= start_date]

        if end_date:
            results = results[results.index.get_level_values("Date") <= end_date]

        fields = results.index.get_level_values("Field").unique()
        kwargs = dict(
            trim_outliers=trim_outliers,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window
        )
        kwargs["returns"] = results.loc["Return"]

        if "Pnl" in fields:
            kwargs["pnl"] = results.loc["Pnl"]
        if "NetExposure" in fields:
            kwargs["net_exposures"] = results.loc["NetExposure"]
        if "AbsExposure" in fields:
            kwargs["abs_exposures"] = results.loc["AbsExposure"]
        if "TotalHoldings" in fields:
            kwargs["total_holdings"] = results.loc["TotalHoldings"]
        if "Turnover" in fields:
            kwargs["turnover"] = results.loc["Turnover"]
        if "Commission" in fields:
            kwargs["commissions"] = results.loc["Commission"]
        if "CommissionAmount" in fields:
            kwargs["commission_amounts"] = results.loc["CommissionAmount"].astype(np.float64)
        if "Slippage" in fields:
            kwargs["slippages"] = results.loc["Slippage"]
        if "Benchmark" in fields:
            kwargs["benchmark"] = results.loc["Benchmark"]

        return cls(**kwargs)

    @classmethod
    def from_moonshot_csv(cls, filepath_or_buffer,
                          start_date=None,
                          end_date=None,
                          trim_outliers=None,
                          how_to_aggregate=None,
                          riskfree=0,
                          compound=True,
                          rolling_sharpe_window=200):
        """
        Creates a DailyPerformance instance from a Moonshot backtest results CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        start_date : str (YYYY-MM-DD), optional
            truncate at this start date (otherwise include entire date range)

        end_date : str (YYYY-MM-DD), optional
            truncate at this end date (otherwise include entire date range)

        trim_outliers: int or float, optional
            discard returns that are more than this many standard deviations from
            the mean. Useful for dealing with data anomalies that cause large
            spikes in plots.

        how_to_aggregate : dict, optional
            a dict of {fieldname: aggregation method} specifying how to aggregate
            fields from intraday to daily. See the docstring for
            `moonchart.utils.intraday_to_daily` for more details.

        riskfree : float, optional
            the riskfree rate (default 0)

        compound : bool
             True for compound/geometric returns, False for arithmetic returns (default True)

        rolling_sharpe_window : int, optional
            compute rolling Sharpe over this many periods (default 200)

        Returns
        -------
        DailyPerformance

        Examples
        --------
        Plot cumulative returns:

        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> perf.cum_returns.plot()
        """
        try:
            results = read_moonshot_csv(filepath_or_buffer)
        except ValueError as e:
            # "ValueError: 'Date' is not in list" might mean the user passed
            # a paramscan csv by mistake
            if "Date" not in repr(e):
                raise
            results = pd.read_csv(filepath_or_buffer)
            if "StrategyOrDate" in results.columns:
                raise MoonchartError("this is a parameter scan CSV, please use ParamscanTearsheet.from_moonshot_csv")
            else:
                raise

        return cls._from_moonshot_or_pnl(
            results,
            start_date=start_date,
            end_date=end_date,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

    @classmethod
    def from_pnl_csv(cls, filepath_or_buffer,
                     start_date=None,
                     end_date=None,
                     trim_outliers=None,
                     how_to_aggregate=None,
                     riskfree=0,
                     compound=True,
                     rolling_sharpe_window=200):
        """
        Creates a DailyPerformance instance from a PNL CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        start_date : str (YYYY-MM-DD), optional
            truncate at this start date (otherwise include entire date range)

        end_date : str (YYYY-MM-DD), optional
            truncate at this end date (otherwise include entire date range)

        trim_outliers: int or float, optional
            discard returns that are more than this many standard deviations from the mean

        how_to_aggregate : dict, optional
            a dict of {fieldname: aggregation method} specifying how to aggregate
            fields from intraday to daily. See the docstring for
            `moonchart.utils.intraday_to_daily` for more details.

        riskfree : float, optional
            the riskfree rate (default 0)

        compound : bool
             True for compound/geometric returns, False for arithmetic returns (default True)

        rolling_sharpe_window : int, optional
            compute rolling Sharpe over this many periods (default 200)

        Returns
        -------
        DailyPerformance

        Examples
        --------
        Plot cumulative returns:

        >>> perf = DailyPerformance.from_pnl_csv("pnl.csv")
        >>> perf.cum_returns.plot()
        """
        results = read_pnl_csv(filepath_or_buffer)

        return cls._from_moonshot_or_pnl(
            results,
            start_date=start_date,
            end_date=end_date,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

    @property
    def cum_returns(self):

        if self._cum_returns is None:
            self._cum_returns = get_cum_returns(self.returns, compound=self.compound)

        return self._cum_returns

    @property
    def cum_commissions(self):

        if self._cum_commissions is None and self.commissions is not None:
            self._cum_commissions = get_cum_returns(self.commissions, compound=False)

        return self._cum_commissions

    @property
    def cum_commission_amounts(self):

        if self._cum_commission_amounts is None and self.commission_amounts is not None:
            self._cum_commission_amounts = self.commission_amounts.cumsum()

        return self._cum_commission_amounts

    @property
    def cum_slippages(self):

        if self._cum_slippages is None and self.slippages is not None:
            self._cum_slippages = get_cum_returns(self.slippages, compound=False)

        return self._cum_slippages

    @property
    def cagr(self):
        if self._cagr is None:
            self._cagr = get_cagr(self.cum_returns, compound=self.compound)

        return self._cagr

    @property
    def sharpe(self):
        if self._sharpe is None:
            self._sharpe = get_sharpe(self.returns, riskfree=self.riskfree)

        return self._sharpe

    @property
    def rolling_sharpe(self):
        if self._rolling_sharpe is None:
            self._rolling_sharpe = get_rolling_sharpe(
                self.returns,
                window=self.rolling_sharpe_window,
                riskfree=self.riskfree)

        return self._rolling_sharpe

    @property
    def drawdowns(self):
        if self._drawdowns is None:
            self._drawdowns = get_drawdowns(self.cum_returns)

        return self._drawdowns

    @property
    def max_drawdown(self):
        return self.drawdowns.min()

    @property
    def cum_pnl(self):
        if self._cum_pnl is None and self.pnl is not None:
            self._cum_pnl = self.pnl.cumsum()

        return self._cum_pnl

    @property
    def benchmark_returns(self):
        """
        Returns a Series of benchmark returns from the DataFrame of benchmark
        prices, if any. If more than one strategy/column has benchmark
        prices, uses the first to compute returns.
        """
        if self._benchmark_returns is not None:
            return self._benchmark_returns

        if self._benchmark_prices is None:
            return None

        have_benchmarks = self._benchmark_prices.notnull().any(axis=0)
        have_benchmarks = have_benchmarks[have_benchmarks]
        if have_benchmarks.empty:
            return None

        col = have_benchmarks.index[0]
        if len(have_benchmarks.index) > 1:
            import warnings
            warnings.warn("Multiple benchmarks found, only using first ({0})".format(col))

        benchmark_prices = self._benchmark_prices[col]

        self._benchmark_returns = benchmark_prices.pct_change().fillna(0)
        self._benchmark_returns.name = "benchmark"

        return self._benchmark_returns

    @property
    def benchmark_cum_returns(self):

        if self._benchmark_cum_returns is None and self._benchmark_returns is not None:
            self._benchmark_cum_returns = get_cum_returns(self._benchmark_returns, compound=True)

        return self._benchmark_cum_returns

class AggregateDailyPerformance(DailyPerformance):
    """
    Class representing aggregate daily performance.

    Given a DailyPerformance instance containing multi-column (that is,
    multi-strategy or detailed single-strategy) DataFrames, this class
    represents the aggregate performance.

    Parameters
    ----------
    performance : DailyPerformance, required
        daily performance results to aggregate

    trim_outliers: int or float, optional
        discard returns that are more than this many standard deviations from the mean
        (copied from DailyPerformance if omitted)

    riskfree : float, optional
        the riskfree rate (copied from DailyPerformance if omitted)

    compound : bool
         True for compound/geometric returns, False for arithmetic returns (copied from
         DailyPerformance if omitted)

    rolling_sharpe_window : int, optional
        compute rolling Sharpe over this many periods (copied from DailyPerformance if
        omitted)

    Returns
    -------
    AggregatePerformance

    Examples
    --------
    Plot aggregate cumulative returns:

    >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
    >>> agg_perf = AggregateDailyPerformance(perf)
    >>> agg_perf.cum_returns.plot()
    """

    def __init__(self, performance, riskfree=None,
                 compound=None,
                 rolling_sharpe_window=None,
                 trim_outliers=None):

        if riskfree is None:
            riskfree = performance.riskfree
        if compound is None:
            compound = performance.compound
        if rolling_sharpe_window is None:
            rolling_sharpe_window = performance.rolling_sharpe_window
        if trim_outliers is None:
            trim_outliers = performance._trim_outliers

        super(AggregateDailyPerformance, self).__init__(
            performance.returns.sum(axis=1),
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window,
            benchmark=performance._benchmark_prices,
            trim_outliers=trim_outliers
        )
        if performance.pnl is not None:
            self.pnl = performance.pnl.sum(axis=1)

        if performance.commission_amounts is not None:
            self.commission_amounts = performance.commission_amounts.sum(axis=1)

        if performance.commissions is not None:
            self.commissions = performance.commissions.sum(axis=1)

        if performance.slippages is not None:
            self.slippages = performance.slippages.sum(axis=1)

        if performance.net_exposures is not None:
            self.net_exposures = performance.net_exposures.sum(axis=1)

        if performance.abs_exposures is not None:
            self.abs_exposures = performance.abs_exposures.sum(axis=1)

        if performance.total_holdings is not None:
            self.total_holdings = performance.total_holdings.sum(axis=1)

        if performance.turnover is not None:
            self.turnover = performance.turnover.sum(axis=1)



# Copyright 2019 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
import seaborn as sns
import empyrical as ep
import scipy.stats
import warnings
from .base import BaseTearsheet
from .perf import DailyPerformance, AggregateDailyPerformance
from .exceptions import MoonchartError
from .utils import with_baseline, get_cum_returns, get_drawdowns

class ShortfallTearsheet(BaseTearsheet):
    """
    Generate a tear sheet of performance stats and plots highlighting the
    shortfall between simulated or benchmark results and actual results.

    See Also
    --------
    ShortfallTearsheet.from_csvs : Create a shortfall tear sheet from CSVs.
    """

    def __init__(self,
                 *args,
                 labels=("simulated", "actual"),
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.x_label, self.y_label = labels

    @classmethod
    def from_csvs(cls, x_filepath_or_buffer, y_filepath_or_buffer,
                  labels=("simulated", "actual"),
                  start_date=None,
                  end_date=None,
                  largest_n=None,
                  shift_x_positions=0,
                  figsize=None, trim_outliers=None,
                  how_to_aggregate=None,
                  pdf_filename=None, riskfree=0,
                  compound=True, rolling_sharpe_window=200):
        """
        Create a shortfall tear sheet comparing simulated results (or other benchmark results)
        with actual results.

        Parameters
        ----------
        x_filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV containing simulated results or other
            benchmark results, usually a Moonshot backtest CSV

        y_filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV containing actual results, usually
            a PNL CSV from the blotter

        labels : tuple, optional
            labels for the x and y results, default ('simulated', 'actual')

        start_date : str (YYYY-MM-DD), optional
            truncate at this start date (otherwise include entire date range)

        end_date : str (YYYY-MM-DD), optional
            truncate at this end date (otherwise include entire date range)

        largest_n : int, optional
            include a "largest shortfalls" table with this many records highligting
            the dates and specific instruments or strategies (depending on whether
            the columns of the input CSVs are instruments or strategies) with the
            largest magnitude shortfall (positive or negative). Also results in
            additional plots depicting performance exluding the largest shortfalls.
            Default is None, meaning don't include a largest shortfalls table or
            additional plots.

        shift_x_positions : int, optional
            shift positions forward (positive integer) or backward (negative integer)
            in simulated/benchmark results. Can be used to align simulated positions
            with actual positions.

        figsize : tuple (width, height), optional
            (width, height) of matplotlib figure. Default is (16, 12)

        trim_outliers: int or float, optional
            discard returns that are more than this many standard deviations
            from the mean

        how_to_aggregate : dict, optional
            a dict of {fieldname: aggregation method} specifying how to aggregate
            fields from intraday to daily. See the docstring for
            `moonchart.utils.intraday_to_daily` for more details.

        pdf_filename : string, optional
            save tear sheet to this filepath as a PDF instead of displaying

        riskfree : float, optional
            the riskfree rate (default 0)

        compound : bool
            True for compound/geometric returns, False for arithmetic returns.
            Default True

        rolling_sharpe_window : int, optional
            compute rolling Sharpe over this many periods (default 200)

        Returns
        -------
        DataFrame or None
            DataFrame of largest shortfalls if largest_n is not None, otherwise None

        Examples
        --------
        >>> from moonshot import ShortfallTearsheet
        >>> ShortfallTearsheet.from_csvs("backtest.csv", "pnl.csv")
        """

        x_perf = DailyPerformance.from_moonshot_csv(
            x_filepath_or_buffer,
            start_date=start_date,
            end_date=end_date,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

        y_perf = DailyPerformance.from_pnl_csv(
            y_filepath_or_buffer,
            start_date=start_date,
            end_date=end_date,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

        t = cls(labels=labels, figsize=figsize, pdf_filename=pdf_filename)
        return t.create_full_tearsheet(x_perf, y_perf, largest_n=largest_n,
                                       shift_x_positions=shift_x_positions)

    def create_full_tearsheet(self, x_performance, y_performance, largest_n=None,
                              shift_x_positions=0):
        """
        Create a shortfall tear sheet comparing simulated results (or other
        benchmark results) with actual results.

        Parameters
        ----------
        x_performance : DailyPerformance, required
            simulated or benchmark performance results

        y_performance : DailyPerformance, required
            actual performance results

        largest_n : int, optional
            include a "largest shortfalls" table with this many records highligting
            the dates and specific instruments or strategies (depending on whether
            the columns of the input CSVs are instruments or strategies) with the
            largest magnitude shortfall (positive or negative). Also results in
            additional plots depicting performance exluding the largest shortfalls.
            Default is None, meaning don't include a largest shortfalls table or
            additional plots.

        shift_x_positions : int, optional
            shift positions forward (positive integer) or backward (negative integer)
            in simulated/benchmark results. Can be used to align simulated positions
            with actual positions.

        Returns
        -------
        DataFrame or None
            DataFrame of largest shortfalls if largest_n is not None, otherwise None

        See Also
        --------
        ShortfallTearsheet.from_csvs : Create a shortfall tear sheet from CSVs.
        """

        y_agg_performance = AggregateDailyPerformance(y_performance)
        x_agg_performance = AggregateDailyPerformance(x_performance)

        self.create_summary_tearsheet(x_agg_performance, y_agg_performance)

        largest_shortfalls = None
        if largest_n:
            largest_shortfalls = self.create_largest_shortfalls_tearsheet(x_performance, y_performance, largest_n=largest_n)

        # Returns
        self._plot_field(
            x_agg_performance.returns,
            y_agg_performance.returns,
            fig_title="Returns",
            cumplots=True
        )

        if largest_shortfalls is not None:
            largest_shortfalls_x_mask = largest_shortfalls.shortfall.unstack().reindex(
                index=x_performance.returns.index,
                columns=x_performance.returns.columns)
            largest_shortfalls_y_mask = largest_shortfalls.shortfall.unstack().reindex(
                index=y_performance.returns.index,
                columns=y_performance.returns.columns)

            # Returns excluding largest shortfalls
            self._plot_field(
                x_performance.returns.where(largest_shortfalls_x_mask.isnull(), 0).sum(axis=1),
                y_performance.returns.where(largest_shortfalls_y_mask.isnull(), 0).sum(axis=1),
                fig_title="Returns",
                fig_title_suffix=" excluding Top {} Largest Shortfalls".format(largest_n),
                cumplots=True
            )

        # Commissions
        if x_agg_performance.commissions is not None and y_agg_performance.commissions is not None:
            self._plot_field(
                x_agg_performance.commissions,
                y_agg_performance.commissions,
                fig_title="Commissions",
                format_as_pct_decimal_places=2)

        # Total holdings
        if x_agg_performance.total_holdings is not None and y_agg_performance.total_holdings is not None:
            self._plot_field(
                x_agg_performance.total_holdings.shift(shift_x_positions),
                y_agg_performance.total_holdings,
                fig_title="Total Holdings",
                format_as_pct_decimal_places=None)

        # Net exposures
        if x_agg_performance.net_exposures is not None and y_agg_performance.net_exposures is not None:
            self._plot_field(
                x_agg_performance.net_exposures.shift(shift_x_positions),
                y_agg_performance.net_exposures,
                fig_title="Net Exposure")

        # Absolute exposures
        if x_agg_performance.abs_exposures is not None and y_agg_performance.abs_exposures is not None:
            self._plot_field(
                x_agg_performance.abs_exposures.shift(shift_x_positions),
                y_agg_performance.abs_exposures,
                fig_title="Absolute Exposure")

        # Turnover
        if x_agg_performance.turnover is not None and y_agg_performance.turnover is not None:
            self._plot_field(
                x_agg_performance.turnover.shift(shift_x_positions),
                y_agg_performance.turnover,
                fig_title="Turnover")

        self._save_or_show()

        return largest_shortfalls

    def create_largest_shortfalls_tearsheet(self, x_performance, y_performance, largest_n=5):
        """
        Create a "largest shortfalls" table highligting the dates and
        specific instruments or strategies (depending on whether the columns
        of the input returns are instruments or strategies) with the largest
        magnitude shortfall (positive or negative).

        Parameters
        ----------
        x_performance : DailyPerformance, required
            simulated or benchmark performance results

        y_performance : DailyPerformance, required
            actual performance results

        largest_n : int, optional
            the number of comparisons to include in the table.

        Returns
        -------
        DataFrame
            DataFrame of largest shortfalls
        """

        shared_columns = x_performance.returns.columns.intersection(y_performance.returns.columns)
        if shared_columns.empty:
            warnings.warn("skipping largest shortfalls table because {x} and {y} have no shared columns".format(
                x=self.x_label, y=self.y_label))
            return

        x_returns = x_performance.returns.stack()
        y_returns = y_performance.returns.stack()

        x_returns.name = self.x_label
        y_returns.name = self.y_label

        returns = pd.concat([x_returns, y_returns], axis=1).fillna(0)

        returns["shortfall"] = returns[self.y_label] - returns[self.x_label]
        returns["abs_shortfall"] = returns.shortfall.abs()

        returns = returns[returns.abs_shortfall > 0]
        if returns.empty:
            warnings.warn("skipping largest shortfalls table because no shortfall found")
            return None

        largest_shortfalls = returns.sort_values(
            "abs_shortfall", ascending=False).drop(
                "abs_shortfall", axis=1).head(largest_n)

        largest_shortfalls_raw = largest_shortfalls.copy()

        largest_shortfalls.index = largest_shortfalls.index.set_names(["Date", "ConId"])
        # format as pct
        largest_shortfalls = (largest_shortfalls * 100).round(2).astype(str) + "%"

        largest_shortfalls = largest_shortfalls.reset_index()
        largest_shortfalls.insert(
            0,
            "Top {} Largest Shortfalls".format(largest_n),
            largest_shortfalls.Date.dt.strftime("%Y-%m-%d").str.cat(largest_shortfalls.ConId, sep=" "))

        largest_shortfalls = largest_shortfalls.drop(["Date","ConId"], axis=1)

        with sns.axes_style("white", {'axes.linewidth': 0}):

            fig = plt.figure("Top Shortfalls", figsize=(6,6))

            axis = fig.add_subplot(111)
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)

            largest_shortfalls_dict = largest_shortfalls.to_dict("split")

            table = axis.table(
                colWidths=[0.4, 0.2, 0.2, 0.2],
                cellText=[largest_shortfalls_dict["columns"]] + largest_shortfalls_dict["data"],
                loc="center"
            )

            for (row, col), cell in table.get_celld().items():
                if row == 0:
                    cell.set_text_props(fontproperties=FontProperties(weight='bold'))

            table.scale(2, 2)
            table.set_fontsize("large")
            fig.tight_layout()

        return largest_shortfalls_raw

    def _plot_field(self, x, y, fig_title,
                    fig_title_suffix="",
                    format_as_pct_decimal_places=1,
                    cumplots=False):

        y.name = self.y_label
        x.name = self.x_label

        try:
            results = pd.concat((y, x), axis=1, sort=True).fillna(0)
        except TypeError:
            # sort was introduced in pandas 0.23
            results = pd.concat((y, x), axis=1).fillna(0)

        fig = None

        if cumplots:

            cum_results = get_cum_returns(results)
            cum_shortfall = cum_results[self.y_label] - cum_results[self.x_label]

            cum_shortfall = pd.concat([
                cum_shortfall.where(cum_shortfall>0),
                cum_shortfall.where(cum_shortfall<0)
                ], axis=1, keys=["positive", "negative"])

            fig, (ax1, ax2, axis) = plt.subplots(
                3, 1, sharex=True, gridspec_kw={
                    "height_ratios":[0.5,0.25,0.25],
                    'hspace': 0.1},
                num="Cumulative " + fig_title + fig_title_suffix, figsize=self.figsize)
            with_baseline(cum_results).plot(ax=ax1, title="Cumulative " + fig_title + fig_title_suffix)
            self._y_format_at_least_two_decimal_places(ax1)
            ax1.set_ylabel("Cumulative " + fig_title)
            ax1.set_xlabel("")

            ax2.set_ylabel("Shortfall")
            self._y_format_as_percentage(ax2, max_decimal_places=2)
            ax2.set_ylabel("Percentage of capital")
            with_baseline(cum_shortfall, value=0).plot(
                ax=ax2, title="Cumulative Shortfall", kind="area",
                stacked=False,
                legend=False,
                color=["C1","C2"] # 2nd and 3rd color in palette are green and red
            )
            ax2.set_xlabel("")

        if fig is None:
            width, height = self.figsize
            fig = plt.figure(fig_title + fig_title_suffix, figsize=(width, height/4))
            axis = fig.add_subplot(111)

        if format_as_pct_decimal_places is not None:
            self._y_format_as_percentage(axis, max_decimal_places=format_as_pct_decimal_places)
            axis.set_ylabel("Percentage of capital")

        shortfall = results[self.y_label] - results[self.x_label]
        results["shortfall"] = shortfall
        results = with_baseline(results, value=0)
        results.plot(ax=axis, title="Daily " + fig_title)
        axis.set_xlabel("")

    def create_summary_tearsheet(self, x_performance, y_performance):
        """
        Create a tear sheet of summary shortfall stats in a table.

        Parameters
        ----------
        x_performance : AggregateDailyPerformance, required
            simulated or benchmark aggregate performance results

        y_performance : AggregateDailyPerformance, required
            actual aggregate performance results

        Returns
        -------
        None

        """
        stats = []

        stats.append(
            ["Start Date",
             [x_performance.returns.index.min().date().isoformat(),
              y_performance.returns.index.min().date().isoformat()]])
        stats.append(
            ["End Date",
             [x_performance.returns.index.max().date().isoformat(),
              y_performance.returns.index.max().date().isoformat()]])

        stats.append(
            ['Total Months',
             [round((x_performance.returns.index.max() - x_performance.returns.index.min()) / pd.Timedelta(1, 'M')),
              round((y_performance.returns.index.max() - y_performance.returns.index.min()) / pd.Timedelta(1, 'M'))]])

        stats.append([" Risk and Returns", ["", ""]])
        stats.append([
            "CAGR",
            ["{0}%".format(round(x_performance.cagr * 100, 1)),
             "{0}%".format(round(y_performance.cagr * 100, 1))]])
        stats.append([
            "Sharpe Ratio",
            ['%.2f' % x_performance.sharpe,
             '%.2f' % y_performance.sharpe]])
        stats.append([
            "Max Drawdown",
            ["{0}%".format(round(x_performance.max_drawdown * 100, 1)),
             "{0}%".format(round(y_performance.max_drawdown * 100, 1))]])
        stats.append([
            "Cumulative Return",
            ["{0}%".format(round(ep.cum_returns_final(x_performance.returns) * 100, 1)),
             "{0}%".format(round(ep.cum_returns_final(y_performance.returns) * 100, 1))]])
        if x_performance.commissions is not None and y_performance.commissions is not None:
            stats.append([
                "Cumulative Commissions",
                ["{0}%".format(round(ep.cum_returns_final(x_performance.commissions) * 100, 1)),
                 "{0}%".format(round(ep.cum_returns_final(y_performance.commissions) * 100, 1))]])
        stats.append([
            "Annual Volatility",
            ["{0}%".format(round(ep.annual_volatility(x_performance.returns) * 100, 1)),
             "{0}%".format(round(ep.annual_volatility(y_performance.returns) * 100, 1))]])
        stats.append([
            "Sortino Ratio",
            ['%.2f' % ep.sortino_ratio(x_performance.returns),
             '%.2f' % ep.sortino_ratio(y_performance.returns)]])
        stats.append([
            "Calmar Ratio",
            ['%.2f' % ep.calmar_ratio(x_performance.returns),
             '%.2f' % ep.calmar_ratio(y_performance.returns)]])
        stats.append([
            "Skew",
            ['%.2f' % scipy.stats.skew(x_performance.returns),
             '%.2f' % scipy.stats.skew(y_performance.returns)]])
        stats.append([
            "Kurtosis",
            ['%.2f' % scipy.stats.kurtosis(x_performance.returns),
             '%.2f' % scipy.stats.kurtosis(y_performance.returns)]])

        stats.append([" Positions and Exposure", ["", ""]])

        if x_performance.abs_exposures is not None and y_performance.abs_exposures is not None:
            stats.append([
                "Absolute Exposure (percentage of capital)",
                ["{0}%".format(round(x_performance.abs_exposures.mean() * 100, 1)),
                 "{0}%".format(round(y_performance.abs_exposures.mean() * 100, 1))]])

        if x_performance.net_exposures is not None and y_performance.net_exposures is not None:
            stats.append([
                "Net Exposure (percentage of capital)",
                ["{0}%".format(round(x_performance.net_exposures.mean() * 100, 1)),
                 "{0}%".format(round(y_performance.net_exposures.mean() * 100, 1))]])

        if x_performance.total_holdings is not None and y_performance.total_holdings is not None:
            stats.append([
                "Average Daily Holdings",
                [round(x_performance.total_holdings.mean()),
                 round(y_performance.total_holdings.mean())]])

        if x_performance.turnover is not None and y_performance.turnover is not None:
            stats.append([
                "Average Daily Turnover (percentage of capital)",
                ["{0}%".format(round(x_performance.turnover.mean() * 100, 1)),
                 "{0}%".format(round(y_performance.turnover.mean() * 100, 1))]])

        if x_performance.abs_exposures is not None and y_performance.abs_exposures is not None:
            stats.append([
                "Normalized CAGR (CAGR/Absolute Exposure)",
                ["{0}%".format(round((x_performance.cagr / x_performance.abs_exposures.mean()) * 100, 1)),
                 "{0}%".format(round((y_performance.cagr / y_performance.abs_exposures.mean()) * 100, 1))]])

        with sns.axes_style("white", {'axes.linewidth': 0}):

            fig = plt.figure("Performance Summary", figsize=(6,6))

            axis = fig.add_subplot(111)
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)

            headings, values = zip(*stats)

            table = axis.table(
                cellText=values,
                rowLabels=headings,
                colLabels=[self.x_label, self.y_label],
                loc="center"
            )
            for (row, col), cell in table.get_celld().items():
                txt = cell.get_text().get_text()
                if row == 0 or txt.startswith(" "):
                    cell.set_text_props(fontproperties=FontProperties(weight='bold'))

            table.scale(1, 2)
            table.set_fontsize("large")
            fig.tight_layout()


# Copyright 2019 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
from .exceptions import MoonchartError

def get_zscores(returns):
    """
    Returns the Z-scores of the input returns.

    Parameters
    ----------
    returns : Series or DataFrame, required
        Series or DataFrame of returns

    Returns
    -------
    Series or DataFrame
    """
    # Ignore 0 returns in calculating z score
    nonzero_returns = returns.where(returns != 0)
    z_scores = (nonzero_returns - nonzero_returns.mean())/nonzero_returns.std()
    return z_scores

def trim_outliers(returns, z_score):
    """
    Zeroes out observations that are too many standard deviations from the
    mean.

    Parameters
    ----------
    returns : Series or DataFrame, required
        Series or DataFrame of returns

    z_score : int or float, required
        maximum standard deviation values are allowed to be from the mean

    Returns
    -------
    Series or DataFrame
    """
    z_scores = get_zscores(returns)
    return returns.where(z_scores.abs() <= z_score, 0)

def with_baseline(data, value=1):
    """
    Prepends a date-indexed Series or DataFrame with an initial row that is
    one period earlier than the first row and has the specified value.

    The typical use case is for generating plots: without a baseline row, a cumulative
    returns plot won't start from 1 if the first day's return is nonzero.

    Parameters
    ----------
    data : Series or DataFrame, required
        Series or DataFrame (for example, of returns)

    value : required
        value to insert in the baseline row

    Returns
    -------
    Series or DataFrame

    Examples
    --------
    Typical usage:

    >>> with_baseline(cum_returns).plot()

    Under the hood:

    >>> cum_returns.head()
        2019-01-02   1.01
        2019-01-03  0.995
    >>> with_baseline(cum_returns)
        2019-01-01      1
        2019-01-02   1.01
        2019-01-03  0.995
    """
    period_length = data.index[1] - data.index[0]
    prior_period = data.index[0] - period_length
    if isinstance(data, pd.DataFrame):
        baseline_row = pd.DataFrame(value, index=[prior_period], columns=data.columns)
    else:
        baseline_row = pd.Series(value, index=[prior_period], name=data.name)
    try:
        data_with_baseline = pd.concat((baseline_row, data), sort=False)
    except TypeError:
        # sort was introduced in pandas 0.23
        data_with_baseline = pd.concat((baseline_row, data))
    return data_with_baseline

def _pad_returns(returns):
    """
    Pads a returns Series or DataFrame with business days, in case the
    existing Date index is sparse (as with PNL csvs). Sparse indexes if not
    padded will affect the Sharpe ratio because the 0 return days will not be
    included in the mean and std.
    """
    bdays = pd.date_range(start=returns.index.min(), end=returns.index.max(),freq="B")
    idx = returns.index.union(bdays)
    return returns.reindex(index=idx).fillna(0)

def _get_sharpe(returns, riskfree=0):
    """
    Private function that returns the Sharpe ratio of the returns. Returns
    should already be padded when calling this function.
    """
    mean = (returns - riskfree).mean()
    if isinstance(mean, float) and mean == 0:
        return 0
    std = (returns - riskfree).std()
    # Returns are assumed to represent daily returns, so annualize the Sharpe ratio
    return mean/std * np.sqrt(252)

def get_sharpe(returns, riskfree=0):
    """
    Returns the Sharpe ratio of the returns.

    Parameters
    ----------
    returns : Series or DataFrame, required
        a Series or DataFrame of returns

    riskfree : float, optional
        the risk-free rate (default 0)

    Returns
    -------
    float or Series of floats
    """
    returns = _pad_returns(returns)
    return _get_sharpe(returns)

def get_rolling_sharpe(returns, window, riskfree=0):
    """
    Computes rolling Sharpe ratios for the returns.

    Parameters
    ----------
    returns : Series or DataFrame, required
        a Series or DataFrame of returns

    window : int, required
        rolling window length

    riskfree : float, optional
        the risk-free rate (default 0)

    Returns
    -------
    Series or DataFrame
    """
    returns = _pad_returns(returns)
    rolling_returns = returns.rolling(window, min_periods=window)
    try:
        return rolling_returns.apply(_get_sharpe, raw=True, kwargs=dict(riskfree=riskfree))
    except TypeError as e:
        # handle pandas<0.23
        if "apply() got an unexpected keyword argument 'raw'" in repr(e):
            return rolling_returns.apply(_get_sharpe, kwargs=dict(riskfree=riskfree))
        else:
            raise

def get_cum_returns(returns, compound=True):
    """
    Computes the cumulative returns of the provided returns.

    Parameters
    ----------
    returns : Series or DataFrame, required
        a Series or DataFrame of returns

    compound : bool
        True for compounded (geometric) returns, False for arithmetic
        returns (default True)

    Returns
    -------
    Series or DataFrame
    """
    if compound:
        cum_returns = (1 + returns).cumprod()
    else:
        cum_returns = returns.cumsum() + 1

    cum_returns.index.name = "Date"
    return cum_returns

def get_cagr(cum_returns, compound=True):
    """
    Computes the CAGR from the cumulative returns.

    Parameters
    ----------
    cum_returns : Series or DataFrame, required
        a Series or DataFrame of cumulative returns

    compound : bool
        compute compound annual growth rate if True, otherwise
        compute average annual return (default True)

    Returns
    -------
    float or Series of floats
    """
    # For DataFrames, apply this function to each Series.
    if isinstance(cum_returns, pd.DataFrame):
        return cum_returns.apply(get_cagr, axis=0)

    # Ignore nulls when compting CAGR
    cum_returns = cum_returns[cum_returns.notnull()]

    if cum_returns.empty:
        return 0

    # Compute the CAGR of the Series
    min_date = cum_returns.index.min()
    max_date = cum_returns.index.max()
    years = ((max_date - min_date).days or 1)/365.0
    ending_value = cum_returns.iloc[-1]
    # Since we are computing CAGR on cumulative returns, the beginning
    # value is always 1.
    beginning_value = 1
    if compound:
        cagr = (ending_value/beginning_value)**(1/years) - 1
    else:
        # Compound annual growth rate doesn't apply to arithmetic
        # returns, so just divide the cum_returns by the number of years
        # to get the annual return
        cagr = (ending_value/beginning_value - 1)/years

    return cagr

def get_drawdowns(cum_returns):
    """
    Computes the drawdowns of the cumulative returns.

    Parameters
    ----------
    cum_returns : Series or DataFrame, required
        a Series or DataFrame of cumulative returns

    Returns
    -------
    Series or DataFrame
    """
    cum_returns = cum_returns[cum_returns.notnull()]
    highwater_marks = cum_returns.expanding().max()
    drawdowns = cum_returns/highwater_marks - 1
    return drawdowns

def get_top_movers(returns, n=10):
    """
    Returns the biggest gainers and losers in the returns.

    Parameters
    ----------
    returns : Series or DataFrame, required
        a Series or DataFrame of returns

    n : int, optional
        the number of biggest gainers and losers to return (default 10)

    Returns
    -------
    Series or DataFrame
    """

    if isinstance(returns, pd.DataFrame):
        returns = returns.stack()

    returns = returns.sort_values()

    try:
        top_movers = pd.concat((returns.head(n), returns.tail(n)), sort=True)
    except TypeError:
        # sort was introduced in pandas 0.23
        top_movers = pd.concat((returns.head(n), returns.tail(n)))

    return top_movers

def intraday_to_daily(results, how=None):
    """
    Roll up a DataFrame of intraday performance results to daily, dropping
    the "Time" level from the multi-index.

    The following aggregation methods are supported:

    extreme: min or max of day, whichever is of greater absolute magnitude
    last: last value of day
    max: max of day
    mean: mean of day
    sum: sum of day

    By default, supported fields are aggregated as follows:

    AbsExposure: max
    AbsWeight: max
    Benchmark: last
    Commission: sum
    CommissionAmount: sum
    NetExposure: extreme
    NetLiquidation: mean
    Pnl: sum
    PositionQuantity: extreme
    PositionValue: extreme
    Price: last
    Return: sum
    Slippage: sum
    TotalHoldings: max
    Turnover: sum
    Weight: extreme

    This can be overridden with the `how` parameter.

    Parameters
    ----------
    results : DataFrame, required
         a DataFrame of intraday Moonshot backtest results or PNL results, with
         a "Time" level in the index

    how : dict, optional
        a dict of {fieldname: aggregation method} specifying how to aggregate
        fields. This is combined with and overrides the default methods.

    Returns
    -------
    DataFrame
        a DataFrame of daily results, without a "Time" level in the index

    Examples
    --------
    Convert intraday Moonshot results to daily:

    >>> intraday_results = read_moonshot_csv("moonshot_intraday_backtest.csv")
    >>> daily_results = intraday_to_daily(intraday_results)
    """
    if "Time" not in results.index.names:
        raise MoonchartError("results DataFrame must have 'Time' level in index")

    fields_in_results = results.index.get_level_values("Field").unique()

    daily_results = {}

    # how to aggregate by field
    field_hows = {
        'AbsExposure': 'max',
        'AbsWeight': 'max',
        'Benchmark': 'last',
        'Commission': 'sum',
        'CommissionAmount': 'sum',
        'NetExposure': 'extreme',
        'NetLiquidation': 'mean',
        'Pnl': 'sum',
        'PositionQuantity': 'extreme',
        'PositionValue': 'extreme',
        'Price': 'last' ,
        'Return': 'sum',
        'Slippage': 'sum',
        'TotalHoldings': 'max',
        'Turnover': 'sum',
        'Weight': 'extreme',
    }

    if how:
        field_hows.update(how)

    for field in fields_in_results:
        if field not in field_hows:
            continue

        field_how = field_hows[field]

        field_results = results.loc[field].astype(pd.np.float64)
        grouped = field_results.groupby(field_results.index.get_level_values("Date"))

        if field_how == "extreme":
            mins = field_results.groupby(field_results.index.get_level_values("Date")).min()
            maxes = field_results.groupby(field_results.index.get_level_values("Date")).max()
            daily_results[field] = mins.where(mins.abs()>maxes.abs(), maxes)
        else:
            daily_results[field] = getattr(grouped, field_how)()

    daily_results = pd.concat(daily_results, names=["Field","Date"])
    return daily_results


# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib import cycler
from matplotlib.backends.backend_pdf import PdfPages
from .utils import (
    with_baseline,
    get_cum_returns)
from matplotlib.ticker import FuncFormatter

# Set seaborn default style
sns.set()

class BaseTearsheet(object):
    """
    Base class for tear sheets.

    Parameters
    ----------
    figsize : tuple (width, height), optional
        (width, height) of matplotlib figure. Default is (16, 12)

    pdf_filename : string, optional
        save tear sheet to this filepath as a PDF instead of displaying

    max_cols_for_details : int, optional
        suppress detailed plots if there are more than this many columns
        (i.e. strategies or securities). Too many plots may cause slow
        rendering. Default 25.
    """
    def __init__(self, pdf_filename=None, figsize=None, max_cols_for_details=25):
        self.figsize = figsize or (16.0, 12.0) # width, height in inches
        plt.rc("axes", axisbelow=True)
        if pdf_filename:
            self._pdf = PdfPages(pdf_filename, keep_empty=True)
        else:
            self._pdf = None

        self._suptitle = None
        self._suptitle_kwargs = {
            "bbox": dict(facecolor="#EAEAF2", edgecolor='white', alpha=0.5)}
        self.max_cols_for_details = max_cols_for_details

    def _save_or_show(self):
        """
        Saves the fig to the multi-page PDF, or shows it.
        """
        if self._pdf:
            for fignum in plt.get_fignums():
                self._pdf.savefig(fignum, bbox_inches='tight')
            plt.close("all")
            self._pdf.close()
        else:
            plt.show()

    def _y_format_as_percentage(self, axis, max_decimal_places=1):
        """
        Sets a Y-axis formatter that converts a decimal to a percentage (e.g.
        0.12 -> 12.0%)
        """
        def format_as_pct(x, pos):
            # Round to max_decimal_places (12.1%) unless it doesn't matter (12%
            # not 12.0%)
            decimal_places = max_decimal_places

            while decimal_places > 0:
                rounded_result = round(x, decimal_places+2)
                more_rounded_result = round(x, decimal_places+1)

                if rounded_result != more_rounded_result:
                    return ('{:.%d%%}' % decimal_places).format(x)

                decimal_places -= 1

            return '{:.0%}'.format(x)

        y_axis_formatter = FuncFormatter(format_as_pct)
        axis.yaxis.set_major_formatter(FuncFormatter(y_axis_formatter))

    def _y_format_at_least_two_decimal_places(self, axis):
        """
        Sets a Y-axis formatter that rounds a decimal to two decimal places.
        """
        def format_at_least_two_decimal_places(x, pos):
            if round(x,2) == round(x,3):
                return '{:.2f}'.format(x)
            else:
                decimal_places = 3
                while True:
                    if math.isclose(round(x, decimal_places), round(x, decimal_places+1)):
                        return round(x, decimal_places)
                    decimal_places += 1

        y_axis_formatter = FuncFormatter(format_at_least_two_decimal_places)
        axis.yaxis.set_major_formatter(FuncFormatter(y_axis_formatter))

    def _get_plot_dimensions(self, plot_count):
        """
        Returns a tuple of rows, cols needed to accomodate the plot_count.
        """
        rows = math.ceil(math.sqrt(plot_count))
        cols = math.ceil(plot_count/rows)
        return rows, cols

    def _clear_legend(self, plot, legend_title=None):
        """
        Anchors the legend to the outside right of the plot area so you can
        see the plot.
        """
        plot.legend(
            loc='center left', bbox_to_anchor=(1, 0.5), title=legend_title)

    def _create_returns_plots(self, performance, subplot, extra_label, figsize=None,
                              legend_title=None):
        """
        Creates agg/details plots for cumulative returns, drawdowns, rolling
        Sharpe, and possibly pnl.
        """
        figsize = figsize or self.figsize

        color_palette = sns.color_palette()

        if isinstance(performance.cum_returns, pd.DataFrame):
            num_series = len(performance.cum_returns.columns)
            if performance.benchmark_returns is not None:
                num_series += 1
            if num_series > 6:
                color_palette = sns.color_palette("hls", num_series)

        with sns.color_palette(color_palette):
            fig = plt.figure("Cumulative Returns", figsize=figsize)
            axis = fig.add_subplot(subplot)
            max_return = performance.cum_returns.max(axis=0)
            if isinstance(max_return, pd.Series):
                max_return = max_return.max(axis=0)
            # If the price more than doubled, use a log scale
            if max_return >= 2:
                axis.set_yscale("log", basey=2)
                axis.set_ylabel("Cumulative return (log scale)")
            else:
                self._y_format_at_least_two_decimal_places(axis)
                axis.set_ylabel("Cumulative return")

            include_commissions = (
                performance.commissions is not None
                # if all commissions are null/0, don't show them
                and (performance.commissions.fillna(0) != 0).any())

            include_slippage = (
                performance.slippages is not None
                # if all slippages are null/0, don't show them
                and (performance.slippages.fillna(0) != 0).any())

            if (
                # a 212 subplot means a detailed plot, which isn't compatible with
                # showing commissions and slippage
                subplot != 212 and (include_commissions or include_slippage)):

                if include_commissions:
                    cum_commissions = performance.cum_commissions
                    cum_commissions.name = "commissions"

                if include_slippage:
                    cum_slippages = performance.cum_slippages
                    cum_slippages.name = "slippage"

                performance.cum_returns.name = "returns"

                cum_gross_returns = performance.cum_returns

                if include_commissions:
                    cum_gross_returns = cum_gross_returns.multiply(cum_commissions)

                if include_slippage:
                    cum_gross_returns = cum_gross_returns.multiply(cum_slippages)

                cum_gross_returns.name = "gross returns"
                breakdown_parts = [performance.cum_returns, cum_gross_returns]

                if include_commissions:
                    breakdown_parts.append(cum_commissions)

                if include_slippage:
                    breakdown_parts.append(cum_slippages)

                try:
                    returns_breakdown = pd.concat(breakdown_parts, axis=1, sort=True)
                except TypeError:
                    # sort was introduced in pandas 0.23
                    returns_breakdown = pd.concat(breakdown_parts, axis=1)
                plot = with_baseline(returns_breakdown).plot(ax=axis, title="Cumulative Returns {0}".format(extra_label))
                axis.set_xlabel("")
                if isinstance(returns_breakdown, pd.DataFrame):
                    self._clear_legend(plot, legend_title=legend_title)
            else:
                plot = with_baseline(performance.cum_returns).plot(ax=axis, title="Cumulative Returns {0}".format(extra_label))
                axis.set_xlabel("")
                if isinstance(performance.cum_returns, pd.DataFrame):
                    self._clear_legend(plot, legend_title=legend_title)

            fig = plt.figure("Drawdowns", figsize=figsize)
            axis = fig.add_subplot(subplot)
            axis.set_ylabel("Drawdown")
            self._y_format_as_percentage(axis)
            plot = performance.drawdowns.plot(ax=axis, title="Drawdowns {0}".format(extra_label))
            axis.set_xlabel("")
            if isinstance(performance.drawdowns, pd.DataFrame):
                self._clear_legend(plot, legend_title=legend_title)

            if performance.cum_pnl is not None:
                # Cumulative pnl will generally be very similar to cumulative
                # return, but they might be different if cumulative return is
                # calcuated based on an account balance that changes due to
                # causes other than the strategies being charted (e.g. other
                # strategies, contributions/withdrawals, etc.)
                fig = plt.figure("Cumulative PNL", figsize=figsize)
                axis = fig.add_subplot(subplot)
                axis.set_ylabel("PNL")
                if (
                    performance.commission_amounts is not None
                    # a 212 subplot means a detailed plot, which isn't compatible with
                    # showing commissions
                    and subplot != 212
                    # if all commissions are null/0, don't show them
                    and (performance.commission_amounts.fillna(0) != 0).any()):
                    cum_commissions = performance.cum_commission_amounts
                    cum_commissions.name = "commissions"
                    cum_pnl = performance.cum_pnl
                    cum_pnl.name = "pnl"
                    cum_gross_pnl = cum_pnl + cum_commissions.abs()
                    cum_gross_pnl.name = "gross pnl"
                    try:
                        pnl_breakdown = pd.concat((cum_pnl, cum_gross_pnl, cum_commissions), axis=1, sort=True)
                    except TypeError:
                        # sort was introduced in pandas 0.23
                        pnl_breakdown = pd.concat((cum_pnl, cum_gross_pnl, cum_commissions), axis=1)
                    plot = pnl_breakdown.plot(ax=axis, title="Cumulative PNL {0}".format(extra_label))
                    axis.set_xlabel("")
                    if isinstance(returns_breakdown, pd.DataFrame):
                        self._clear_legend(plot, legend_title=legend_title)
                else:
                    plot = performance.cum_pnl.plot(ax=axis, title="Cumulative PNL {0}".format(extra_label))
                    axis.set_xlabel("")
                    if isinstance(performance.cum_pnl, pd.DataFrame):
                        self._clear_legend(plot, legend_title=legend_title)

            if len(performance.rolling_sharpe.index) > performance.rolling_sharpe_window:
                fig = plt.figure("Rolling Sharpe", figsize=figsize)
                axis = fig.add_subplot(subplot)
                axis.set_ylabel("Sharpe ratio")
                self._y_format_at_least_two_decimal_places(axis)
                plot = performance.rolling_sharpe.plot(ax=axis, title="Rolling Sharpe ({0}-day) {1}".format(
                    performance.rolling_sharpe_window, extra_label))
                axis.set_xlabel("")
                if isinstance(performance.rolling_sharpe, pd.DataFrame):
                    self._clear_legend(plot, legend_title=legend_title)

            if performance.benchmark_returns is not None:
                fig = plt.figure("Cumulative Returns vs Benchmark", figsize=figsize)
                axis = fig.add_subplot(subplot)
                max_return = performance.cum_returns.max(axis=0)
                if isinstance(max_return, pd.Series):
                    max_return = max_return.max(axis=0)

                # If the price more than doubled, use a log scale
                if max_return >= 2:
                    axis.set_yscale("log", basey=2)
                    axis.set_ylabel("Cumulative return (log scale)")
                else:
                    self._y_format_at_least_two_decimal_places(axis)
                    axis.set_ylabel("Cumulative return")

                if isinstance(performance.cum_returns, pd.Series):
                    performance.cum_returns.name = "strategy"
                benchmark_cum_returns = performance.benchmark_cum_returns
                try:
                    vs_benchmark = pd.concat((performance.cum_returns, benchmark_cum_returns), axis=1, sort=True)
                except TypeError:
                    # sort was introduced in pandas 0.23
                    vs_benchmark = pd.concat((performance.cum_returns, benchmark_cum_returns), axis=1)
                plot = with_baseline(vs_benchmark).plot(ax=axis, title="Cumulative Returns vs Benchmark {0}".format(extra_label))
                axis.set_xlabel("")
                if isinstance(vs_benchmark, pd.DataFrame):
                    self._clear_legend(plot, legend_title=legend_title)



# Copyright 2017 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
import math
import warnings
import empyrical as ep
import scipy.stats
from .perf import DailyPerformance, AggregateDailyPerformance
from .base import BaseTearsheet
from .exceptions import MoonchartError
from .utils import (
    with_baseline,
    get_sharpe,
    get_cagr,
    get_cum_returns,
    get_drawdowns
)

class Tearsheet(BaseTearsheet):
    """
    Generates a tear sheet of performance stats and graphs.
    """
    @classmethod
    def from_moonshot_csv(cls, filepath_or_buffer, figsize=None,
                          max_cols_for_details=25, trim_outliers=None,
                          how_to_aggregate=None,
                          pdf_filename=None, riskfree=0,
                          compound=True, rolling_sharpe_window=200):
        """
        Create a full tear sheet from a moonshot backtest results CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath of CSV or file-like object

        figsize : tuple (width, height), optional
            (width, height) of matplotlib figure. Default is (16, 12)

        max_cols_for_details : int, optional
            suppress detailed plots if there are more than this many columns
            (i.e. strategies or securities). Too many plots may cause slow
            rendering. Default 25.

        trim_outliers: int or float, optional
            discard returns that are more than this many standard deviations
            from the mean. Useful for dealing with data anomalies that cause
            large spikes in plots.

        how_to_aggregate : dict, optional
            a dict of {fieldname: aggregation method} specifying how to aggregate
            fields from intraday to daily. See the docstring for
            `moonchart.utils.intraday_to_daily` for more details.

        pdf_filename : string, optional
            save tear sheet to this filepath as a PDF instead of displaying

        riskfree : float, optional
            the riskfree rate (default 0)

        compound : bool
            True for compound/geometric returns, False for arithmetic returns.
            Default True

        rolling_sharpe_window : int, optional
            compute rolling Sharpe over this many periods (default 200)

        Returns
        -------
        None

        Examples
        --------
        >>> from moonshot import Tearsheet
        >>> Tearsheet.from_moonshot_csv("backtest_results.csv")
        """
        perf = DailyPerformance.from_moonshot_csv(
            filepath_or_buffer,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

        t = cls(figsize=figsize,
                max_cols_for_details=max_cols_for_details,
                pdf_filename=pdf_filename)

        return t.create_full_tearsheet(perf)

    @classmethod
    def from_pnl_csv(cls, filepath_or_buffer, figsize=None,
                     max_cols_for_details=25, trim_outliers=None,
                     how_to_aggregate=None,
                     pdf_filename=None, riskfree=0,
                     compound=True, rolling_sharpe_window=200):
        """
        Create a full tear sheet from a pnl CSV.

        Parameters
        ----------
        filepath_or_buffer : str or file-like object
            filepath or file-like object of the CSV

        figsize : tuple (width, height), optional
            (width, height) of matplotlib figure. Default is (16, 12)

        max_cols_for_details : int, optional
            suppress detailed plots if there are more than this many columns
            (i.e. strategies or securities). Too many plots may cause slow
            rendering. Default 25.

        trim_outliers: int or float, optional
            discard returns that are more than this many standard deviations
            from the mean

        how_to_aggregate : dict, optional
            a dict of {fieldname: aggregation method} specifying how to aggregate
            fields from intraday to daily. See the docstring for
            `moonchart.utils.intraday_to_daily` for more details.

        pdf_filename : string, optional
            save tear sheet to this filepath as a PDF instead of displaying

        riskfree : float, optional
            the riskfree rate (default 0)

        compound : bool
            True for compound/geometric returns, False for arithmetic returns.
            Default True

        rolling_sharpe_window : int, optional
            compute rolling Sharpe over this many periods (default 200)

        Returns
        -------
        None
        """
        perf = DailyPerformance.from_pnl_csv(
            filepath_or_buffer,
            trim_outliers=trim_outliers,
            how_to_aggregate=how_to_aggregate,
            riskfree=riskfree,
            compound=compound,
            rolling_sharpe_window=rolling_sharpe_window)

        t = cls(figsize=figsize,
                max_cols_for_details=max_cols_for_details,
                pdf_filename=pdf_filename)

        return t.create_full_tearsheet(perf)

    def create_full_tearsheet(self, performance):
        """
        Create a full tear sheet of performance results including returns
        plots, returns by year plots, and position-related plots.

        Parameters
        ----------
        performance : instance
            a DailyPerformance instance

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_full_tearsheet(perf)

        See Also
        --------
        Tearsheet.from_moonshot_csv : create a full tear sheet from a Moonshot CSV
        """
        agg_performance = AggregateDailyPerformance(performance)

        num_cols = len(performance.returns.columns)
        if num_cols > self.max_cols_for_details:
            warnings.warn("Suppressing details because there are more than {0} columns "
                          "(you can control this setting by modifying "
                          "Tearsheet.max_cols_for_details)".format(
                              self.max_cols_for_details))

        self.create_summary_tearsheet(performance, agg_performance)
        self.create_returns_tearsheet(performance, agg_performance)
        self.create_returns_by_year_tearsheet(performance, agg_performance)

        if any([exposures is not None for exposures in (
            performance.net_exposures, performance.abs_exposures)]):
            self.create_positions_tearsheet(performance, agg_performance)

        self._create_constituents_tearsheet(performance)

        self._save_or_show()

    def create_summary_tearsheet(self, performance, agg_performance=None):
        """
        Create a tear sheet of summary performance stats in a table.

        Parameters
        ----------
        performance : DailyPerformance, required
            a DailyPerformance instance

        agg_performance : AggregateDailyPerformance, optional
            an AggregateDailyPerformance instance. Constructed from performance
            if not provided.

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_summary_tearsheet(perf)
        """
        if agg_performance is None:
            agg_performance = AggregateDailyPerformance(performance)

        stats = []

        if agg_performance.pnl is not None:
            stats.append(["PNL", round(agg_performance.pnl.sum(), 2)])
        if agg_performance.commission_amounts is not None:
            stats.append(["Commissions", round(agg_performance.commission_amounts.sum(), 2)])

        stats.append(["Start Date", agg_performance.returns.index.min().date().isoformat()])
        stats.append(["End Date", agg_performance.returns.index.max().date().isoformat()])

        stats.append(['Total Months', round(
            (agg_performance.returns.index.max() - agg_performance.returns.index.min()) / pd.Timedelta(1, 'M'))])

        stats.append(["", " Risk and Returns"])
        stats.append(["CAGR", "{0}%".format(round(agg_performance.cagr * 100, 1))])
        stats.append([
            "Sharpe Ratio",
            '%.2f' % agg_performance.sharpe])
        stats.append([
            "Max Drawdown",
            "{0}%".format(round(agg_performance.max_drawdown * 100, 1))])
        stats.append([
            "Cumulative Return",
            "{0}%".format(round(ep.cum_returns_final(agg_performance.returns) * 100, 1))])
        stats.append([
            "Annual Volatility",
            "{0}%".format(round(ep.annual_volatility(agg_performance.returns) * 100, 1))])
        stats.append([
            "Sortino Ratio",
            '%.2f' % ep.sortino_ratio(agg_performance.returns)])
        stats.append([
            "Calmar Ratio",
            '%.2f' % ep.calmar_ratio(agg_performance.returns)])
        stats.append([
            "Skew",
            '%.2f' % scipy.stats.skew(agg_performance.returns)])
        stats.append([
            "Kurtosis",
            '%.2f' % scipy.stats.kurtosis(agg_performance.returns)])

        if any([field is not None for field in (
            agg_performance.abs_exposures,
            agg_performance.net_exposures,
            agg_performance.total_holdings,
            agg_performance.turnover
            )]):
            stats.append(["", " Positions and Exposure"])

        if agg_performance.abs_exposures is not None:
            avg_abs_exposures = agg_performance.abs_exposures.mean()
            stats.append([
                "Absolute Exposure (percentage of capital)",
                "{0}%".format(round(avg_abs_exposures * 100, 1))])

        if agg_performance.net_exposures is not None:
            avg_net_exposures = agg_performance.net_exposures.mean()
            stats.append([
                "Net Exposure (percentage of capital)",
                "{0}%".format(round(avg_net_exposures * 100, 1))])

        if agg_performance.total_holdings is not None:
            avg_daily_holdings = agg_performance.total_holdings.mean()
            stats.append([
                "Average Daily Holdings",
                round(avg_daily_holdings)])

        if agg_performance.turnover is not None:
            avg_daily_turnover = agg_performance.turnover.mean()
            stats.append([
                "Average Daily Turnover (percentage of capital)",
                "{0}%".format(round(avg_daily_turnover * 100, 1))])

        if agg_performance.abs_exposures is not None:
            norm_cagr = agg_performance.cagr / avg_abs_exposures
            stats.append([
                "Normalized CAGR (CAGR/Absolute Exposure)",
                "{0}%".format(round(norm_cagr * 100, 1))])

        with sns.axes_style("white", {'axes.linewidth': 0}):

            fig = plt.figure("Performance Summary", figsize=(6,6))

            axis = fig.add_subplot(111)
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)

            headings, values = zip(*stats)

            table = axis.table(
                cellText=[[v] for v in values],
                rowLabels=headings,
                colLabels=["Performance Summary"],
                loc="center"
            )
            for (row, col), cell in table.get_celld().items():
                txt = cell.get_text().get_text()
                if row == 0 or txt.startswith(" "):
                    cell.set_text_props(fontproperties=FontProperties(weight='bold'))

            table.scale(1, 2)
            table.set_fontsize("large")
            fig.tight_layout()

    def _create_constituents_tearsheet(self, performance):
        """
        Create a tear sheet of the strategies or symbols in the data.
        """
        with sns.axes_style("white", {'axes.linewidth': 0}):

            fig = plt.figure("Strategies or Securities", figsize=(6,6))

            axis = fig.add_subplot(111)
            axis.get_xaxis().set_visible(False)
            axis.get_yaxis().set_visible(False)

            cols = list(performance.returns.columns)
            if len(cols) > 58:
                hidden_cols = len(cols) - 58
                cols = cols[0:58]
                cols.append("and {0} more".format(hidden_cols))

            cells_per_row = 4
            cells = ["Included columns:"] + cols
            num_cells = len(cells)
            if num_cells > cells_per_row and num_cells % cells_per_row != 0:
                # Cells must be divisible by cells_per_row for matplotlib table
                extra_cells_required = cells_per_row - num_cells % cells_per_row
                for _ in range(extra_cells_required):
                    cells.append("")

            table = axis.table(
                cellText=[cells[i:i + cells_per_row] for i in range(0, len(cells), cells_per_row)],
                loc="top"
            )
            for (row, col), cell in table.get_celld().items():
                if (row == 0) and (col == 0):
                    cell.set_text_props(fontproperties=FontProperties(weight='bold'))

            table.scale(2, 2)
            table.set_fontsize("large")
            fig.tight_layout()

    def create_returns_tearsheet(self, performance, agg_performance=None):
        """
        Create a tear sheet of returns-related plots.

        The included plots depend on what is present in the performance data.
        Always plots cumulative returns, drawdowns, and rolling Sharpe. Plots
        cumulative returns vs benchmark if benchmark is present. Plots
        cumulative PNL if PNL is present. For multi-column performance
        data (multi-strategy or detailed single-strategy), plots bar
        charts of Sharpe, CAGR, and PNL if present.

        Parameters
        ----------
        performance : DailyPerformance, required
            a DailyPerformance instance

        agg_performance : AggregateDailyPerformance, optional
            an AggregateDailyPerformance instance. Constructed from performance
            if not provided.

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_returns_tearsheet(perf)
        """
        if agg_performance is None:
            agg_performance = AggregateDailyPerformance(performance)

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details

        width, height = self.figsize
        # cut height in half if not showing details
        if not show_details:
            height /= 2

        self._create_returns_plots(
            agg_performance,
            subplot=211 if show_details else 111,
            extra_label="(Aggregate)" if show_details else "",
            figsize=(width, height))

        if show_details:
            self._create_returns_plots(performance, subplot=212, extra_label="(Details)")
            self._create_detailed_returns_bar_charts(performance)

    def _create_detailed_returns_bar_charts(self, performance):

        fig = plt.figure("Returns (Details)", figsize=self.figsize)

        color_palette = sns.color_palette()
        num_series = len(performance.cum_returns.columns)
        if num_series > 6:
            color_palette = sns.color_palette("hls", num_series)

        with sns.color_palette(color_palette):

            axis = fig.add_subplot(2,2,1)
            axis.set_ylabel("CAGR")
            self._y_format_as_percentage(axis)
            cagr = performance.cagr.copy()
            cagr.index = cagr.index.astype(str).str.wrap(10)
            cagr.plot(ax=axis, kind="bar", title="CAGR (Details)")

            axis = fig.add_subplot(2,2,2)
            self._y_format_at_least_two_decimal_places(axis)
            axis.set_ylabel("Sharpe ratio")
            sharpe = performance.sharpe.copy()
            sharpe.index = sharpe.index.astype(str).str.wrap(10)
            sharpe.plot(ax=axis, kind="bar", title="Sharpe (Details)")

            axis = fig.add_subplot(2,2,3)
            axis.set_ylabel("Drawdown")
            self._y_format_as_percentage(axis)
            max_drawdowns = performance.max_drawdown.copy()
            max_drawdowns.index = max_drawdowns.index.astype(str).str.wrap(10)
            max_drawdowns.plot(ax=axis, kind="bar", title="Max drawdown (Details)")

        fig.tight_layout()

        if performance.pnl is not None:
            fig = plt.figure("PNL (Details)", figsize=self.figsize)
            axis = fig.add_subplot(111)
            axis.set_ylabel("PNL")
            pnl = performance.pnl.sum()
            if performance.commission_amounts is not None:
                pnl.name = "pnl"
                commission_amounts = performance.commission_amounts.sum()
                commission_amounts.name = "commissions"
                gross_pnl = pnl + commission_amounts
                gross_pnl.name = "gross pnl"
                try:
                    pnl = pd.concat((pnl, gross_pnl, commission_amounts), axis=1, sort=True)
                except TypeError:
                    # sort was introduced in pandas 0.23
                    pnl = pd.concat((pnl, gross_pnl, commission_amounts), axis=1)
            pnl.plot(
                ax=axis, kind="bar", title="PNL (Details)")

    def create_positions_tearsheet(self, performance, agg_performance=None):
        """
        Create a tear sheet of position-related plots.

        Includes plots of net and absolute daily exposure, number of daily
        holdings, and daily turnover.

        Parameters
        ----------
        performance : DailyPerformance, required
            a DailyPerformance instance

        agg_performance : AggregateDailyPerformance, optional
            an AggregateDailyPerformance instance. Constructed from performance
            if not provided.

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_positions_tearsheet(perf)
        """
        if agg_performance is None:
            agg_performance = AggregateDailyPerformance(performance)

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details

        width, height = self.figsize
        # cut height in half if not showing details
        if not show_details:
            height /= 2

        self._create_positions_plots(
            agg_performance,
            subplot=211 if show_details else 111,
            extra_label="(Aggregate)" if show_details else "",
            figsize=(width, height))

        if show_details:
            self._create_positions_plots(performance, subplot=212, extra_label="(Details)")
            self._create_detailed_positions_bar_charts(performance)

    def _create_positions_plots(self, performance, subplot, extra_label, figsize=None):

        figsize = figsize or self.figsize

        color_palette = sns.color_palette()

        if isinstance(performance.returns, pd.DataFrame):
            num_series = len(performance.returns.columns)
            if num_series > 6:
                color_palette = sns.color_palette("hls", num_series)

        with sns.color_palette(color_palette):

            if performance.abs_exposures is not None:
                fig = plt.figure("Absolute Exposure", figsize=figsize)
                axis = fig.add_subplot(subplot)
                self._y_format_as_percentage(axis)
                plot = performance.abs_exposures.round(2).plot(ax=axis, title="Absolute Exposure {0}".format(extra_label))
                axis.set_ylabel("Percentage of capital")
                axis.set_xlabel("")
                if isinstance(performance.abs_exposures, pd.DataFrame):
                    self._clear_legend(plot)

            if performance.net_exposures is not None:
                fig = plt.figure("Net Exposure", figsize=figsize)
                axis = fig.add_subplot(subplot)
                self._y_format_as_percentage(axis)
                plot = performance.net_exposures.round(2).plot(ax=axis, title="Net Exposure {0}".format(extra_label))
                axis.set_ylabel("Percentage of capital")
                axis.set_xlabel("")
                if isinstance(performance.net_exposures, pd.DataFrame):
                    self._clear_legend(plot)

            if performance.total_holdings is not None:
                fig = plt.figure("Daily Holdings", figsize=figsize)
                axis = fig.add_subplot(subplot)
                plot = performance.total_holdings.plot(ax=axis, title="Daily Holdings {0}".format(extra_label))
                axis.set_ylabel("Number of holdings")
                axis.set_xlabel("")
                if isinstance(performance.total_holdings, pd.DataFrame):
                    self._clear_legend(plot)

            if performance.turnover is not None:
                fig = plt.figure("Daily Turnover", figsize=figsize)
                axis = fig.add_subplot(subplot)
                self._y_format_as_percentage(axis)
                turnover = performance.turnover
                plot = turnover.plot(ax=axis, title="Daily Turnover {0}".format(extra_label))
                axis.set_ylabel("Percentage of capital")
                axis.set_xlabel("")
                if isinstance(turnover, pd.DataFrame):
                    self._clear_legend(plot)

    def _create_detailed_positions_bar_charts(self, performance):

        # extend figsize due to 3 rows
        width, height = self.figsize
        figsize = width, height*1.5

        fig = plt.figure("Positions (Details)", figsize=figsize)

        color_palette = sns.color_palette()
        num_series = len(performance.cum_returns.columns)
        if num_series > 6:
            color_palette = sns.color_palette("hls", num_series)

        total_plots = sum([1 for field in (
            performance.abs_exposures,
            performance.net_exposures,
            performance.total_holdings,
            performance.turnover,
            performance.abs_exposures) if field is not None])

        rows = math.ceil(total_plots/2)

        with sns.color_palette(color_palette):

            next_pos = 1

            if performance.abs_exposures is not None:
                avg_abs_exposures = performance.abs_exposures.mean()
                axis = fig.add_subplot(rows,2,next_pos)
                next_pos += 1
                self._y_format_as_percentage(axis)
                avg_abs_exposures.plot(ax=axis, kind="bar", title="Avg Absolute Exposure (Details)")
                axis.set_ylabel("Percentage of capital")

            if performance.net_exposures is not None:
                avg_net_exposures = performance.net_exposures.mean()
                axis = fig.add_subplot(rows,2,next_pos)
                next_pos += 1
                self._y_format_as_percentage(axis)
                avg_net_exposures.plot(ax=axis, kind="bar", title="Avg Net Exposure (Details)")
                axis.set_ylabel("Percentage of capital")

            if performance.total_holdings is not None:
                avg_total_holdings = performance.total_holdings.mean()
                axis = fig.add_subplot(rows,2,next_pos)
                next_pos += 1
                avg_total_holdings.plot(ax=axis, kind="bar", title="Avg Daily Holdings (Details)")
                axis.set_ylabel("Number of holdings")

            if performance.turnover is not None:
                avg_daily_turnover = performance.turnover.mean()
                axis = fig.add_subplot(rows,2,next_pos)
                next_pos += 1
                self._y_format_as_percentage(axis)
                avg_daily_turnover.plot(ax=axis, kind="bar", title="Avg Daily Turnover (Details)")
                axis.set_ylabel("Percentage of capital")

            if performance.abs_exposures is not None:
                norm_cagrs = performance.cagr / avg_abs_exposures
                axis = fig.add_subplot(rows,2,next_pos)
                next_pos += 1
                self._y_format_as_percentage(axis)
                norm_cagrs.plot(ax=axis, kind="bar", title="Normalized CAGR (CAGR/Absolute Exposure) (Details)")
                axis.set_ylabel("Normalized CAGR")

        fig.tight_layout()

    def create_returns_by_year_tearsheet(self, performance, agg_performance=None):
        """
        Plots bar charts showing CAGR and Sharpe by year.

        Parameters
        ----------
        performance : DailyPerformance, required
            a DailyPerformance instance

        agg_performance : AggregateDailyPerformance, optional
            an AggregateDailyPerformance instance. Constructed from performance
            if not provided.

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_returns_by_year_tearsheet(perf)
        """
        if agg_performance is None:
            agg_performance = AggregateDailyPerformance(performance)

        num_cols = len(performance.returns.columns)
        show_details = num_cols > 1 and num_cols <= self.max_cols_for_details

        width, height = self.figsize
        # cut height in half if not showing details
        if not show_details:
            height /= 2

        fig = plt.figure("Returns by Year", figsize=(width, height))

        self._create_returns_by_year_plots(
            agg_performance,
            rows=2 if show_details else 1,
            row=1,
            fig=fig,
            extra_label="(Aggregate)" if show_details else "")

        if show_details:
            self._create_returns_by_year_plots(performance, rows=2, row=2, fig=fig, extra_label="(Details)")

    def _create_returns_by_year_plots(self, performance, rows, row, fig, extra_label):

        color_palette = sns.color_palette()

        if isinstance(performance.returns, pd.DataFrame):
            num_series = len(performance.cum_returns.columns)
            if num_series > 6:
                color_palette = sns.color_palette("hls", num_series)
        else:
            color_palette = sns.color_palette()[0:1]

        grouped_returns = performance.returns.groupby(performance.returns.index.year)
        cagrs_by_year = grouped_returns.apply(lambda x: get_cagr(
            get_cum_returns(x, compound=performance.compound),
            compound=performance.compound))
        sharpes_by_year = grouped_returns.apply(get_sharpe, riskfree=performance.riskfree)

        cols = 2
        # 2 cols per row, minus 1, gives the start position
        start_at = 2 * row - 1
        with sns.color_palette(color_palette):
            axis = fig.add_subplot(rows, 2, start_at)
            axis.set_ylabel("CAGR")
            self._y_format_as_percentage(axis)
            plot = cagrs_by_year.plot(ax=axis, kind="bar", title="CAGR by Year {0}".format(extra_label))
            axis.set_xlabel("")
            if isinstance(cagrs_by_year, pd.DataFrame):
                # Remove legend, rely on legend from Sharpe plot
                plot.legend_.remove()

            axis = fig.add_subplot(rows, 2, start_at+1)
            axis.set_ylabel("Sharpe ratio")
            self._y_format_at_least_two_decimal_places(axis)
            plot = sharpes_by_year.plot(ax=axis, kind="bar", title="Sharpe by Year {0}".format(extra_label))
            axis.set_xlabel("")
            if isinstance(sharpes_by_year, pd.DataFrame):
                self._clear_legend(plot)

        fig.tight_layout()

    def create_montecarlo_tearsheet(self, performance, cycles=5, aggregate_before_shuffle=True):
        """
        Run a Montecarlo simulation by shuffling the DataFrame of returns a specified
        number of times and plotting the shuffled returns against the original returns.

        Parameters
        ----------
        performance : DailyPerformance, required
            a DailyPerformance instance

        cycles : int, optional
            the number of Montecarlo simulations (default 5)

        aggregate_before_shuffle : bool
            whether to aggregate daily returns before or after shuffling. Only relevant to
            multi-column (that is, multi-strategy or detailed single-strategy) DataFrames.
            If True, aggregated daily returns are preserved and only the order of days is
            randomized. If False, each column's returns are shuffled separately, without
            preservation of daily aggregations. False is more random. True may be preferable
            if daily returns across columns are expected to be correlated. Default True.

        Returns
        -------
        None

        Examples
        --------
        >>> from moonchart import DailyPerformance, Tearsheet
        >>> perf = DailyPerformance.from_moonshot_csv("backtest_results.csv")
        >>> t = Tearsheet()
        >>> t.create_montecarlo_tearsheet(perf, cycles=10)
        """
        all_simulations = []

        returns = performance.returns

        if aggregate_before_shuffle:
            returns = returns.sum(axis=1)

        for i in range(cycles):
            if aggregate_before_shuffle:
                sim_returns = pd.Series(np.random.permutation(returns), index=returns.index)
            else:
                sim_returns = returns.apply(np.random.permutation).sum(axis=1)
            all_simulations.append(sim_returns)

        try:
            sim_returns = pd.concat(all_simulations, axis=1, sort=False)
        except TypeError:
            # sort was introduced in pandas 0.23
            sim_returns = pd.concat(all_simulations, axis=1)

        if not aggregate_before_shuffle:
            returns = returns.sum(axis=1)

        cum_sim_returns = get_cum_returns(sim_returns, compound=performance.compound)
        cum_returns = get_cum_returns(returns, compound=performance.compound)
        fig = plt.figure("Montecarlo Simulation", figsize=self.figsize)
        fig.suptitle(self._suptitle, **self._suptitle_kwargs)
        axis = fig.add_subplot(211)
        self._y_format_at_least_two_decimal_places(axis)
        axis.set_ylabel("Cumulative return")
        with_baseline(cum_sim_returns).plot(ax=axis, title="Montecarlo Cumulative Returns ({0} cycles)".format(cycles), legend=False)
        with_baseline(cum_returns).plot(ax=axis, linewidth=4, color="black")
        axis = fig.add_subplot(212)
        self._y_format_as_percentage(axis)
        axis.set_ylabel("Drawdown")
        get_drawdowns(cum_sim_returns).plot(ax=axis, title="Montecarlo Drawdowns ({0} cycles)".format(cycles), legend=False)
        get_drawdowns(cum_returns).plot(ax=axis, linewidth=4, color="black")




from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .tearsheet import Tearsheet
from .perf import DailyPerformance, AggregateDailyPerformance
from .paramscan import ParamscanTearsheet
from .shortfall import ShortfallTearsheet